"""
Set up the version of Salt
"""

import argparse
import operator
import os
import platform
import re
import sys
from collections import namedtuple
from functools import total_ordering

MAX_SIZE = sys.maxsize
VERSION_LIMIT = MAX_SIZE - 200


# ----- ATTENTION --------------------------------------------------------------------------------------------------->
#
# ALL major version bumps, new release codenames, MUST be defined in the SaltStackVersion.NAMES dictionary, i.e.:
#
#    class SaltStackVersion:
#
#        NAMES = {
#            'Hydrogen': (2014, 1),   # <- This is the tuple to bump versions
#            ( ... )
#        }
#
#
# ONLY UPDATE CODENAMES AFTER BRANCHING
#
# As an example, The Helium codename must only be properly defined with "(2014, 7)" after Hydrogen, "(2014, 1)", has
# been branched out into its own branch.
#
# ALL OTHER VERSION INFORMATION IS EXTRACTED FROM THE GIT TAGS
#
# <---- ATTENTION ----------------------------------------------------------------------------------------------------


@total_ordering
class SaltVersion(namedtuple("SaltVersion", "name, info, released")):
    __slots__ = ()

    def __new__(cls, name, info, released=False):
        if isinstance(info, int):
            info = (info,)
        return super().__new__(cls, name, info, released)

    def __eq__(self, other):
        return self.info == other.info

    def __gt__(self, other):
        return self.info > other.info


class SaltVersionsInfo(type):

    _sorted_versions = ()
    _current_release = None
    _previous_release = None
    _next_release = None

    # pylint: disable=bad-whitespace
    # ----- Please refrain from fixing whitespace ---------------------------------->
    # The idea is to keep this readable.
    # -------------------------------------------------------------------------------
    # fmt: off
    HYDROGEN      = SaltVersion("Hydrogen"     , info=(2014, 1),  released=True)
    HELIUM        = SaltVersion("Helium"       , info=(2014, 7),  released=True)
    LITHIUM       = SaltVersion("Lithium"      , info=(2015, 5),  released=True)
    BERYLLIUM     = SaltVersion("Beryllium"    , info=(2015, 8),  released=True)
    BORON         = SaltVersion("Boron"        , info=(2016, 3),  released=True)
    CARBON        = SaltVersion("Carbon"       , info=(2016, 11), released=True)
    NITROGEN      = SaltVersion("Nitrogen"     , info=(2017, 7),  released=True)
    OXYGEN        = SaltVersion("Oxygen"       , info=(2018, 3),  released=True)
    FLUORINE      = SaltVersion("Fluorine"     , info=(2019, 2),  released=True)
    NEON          = SaltVersion("Neon"         , info=3000,       released=True)
    SODIUM        = SaltVersion("Sodium"       , info=3001,       released=True)
    MAGNESIUM     = SaltVersion("Magnesium"    , info=3002,       released=True)
    ALUMINIUM     = SaltVersion("Aluminium"    , info=3003,       released=True)
    SILICON       = SaltVersion("Silicon"      , info=3004,       released=True)
    PHOSPHORUS    = SaltVersion("Phosphorus"   , info=3005,       released=True)
    SULFUR        = SaltVersion("Sulfur"       , info=3006,       released=True)
    CHLORINE      = SaltVersion("Chlorine"     , info=3007,       released=True)
    ARGON         = SaltVersion("Argon"        , info=3008)
    POTASSIUM     = SaltVersion("Potassium"    , info=3009)
    CALCIUM       = SaltVersion("Calcium"      , info=3010)
    SCANDIUM      = SaltVersion("Scandium"     , info=3011)
    TITANIUM      = SaltVersion("Titanium"     , info=3012)
    VANADIUM      = SaltVersion("Vanadium"     , info=3013)
    CHROMIUM      = SaltVersion("Chromium"     , info=3014)
    MANGANESE     = SaltVersion("Manganese"    , info=3015)
    IRON          = SaltVersion("Iron"         , info=3016)
    COBALT        = SaltVersion("Cobalt"       , info=3017)
    NICKEL        = SaltVersion("Nickel"       , info=3018)
    COPPER        = SaltVersion("Copper"       , info=3019)
    ZINC          = SaltVersion("Zinc"         , info=3020)
    GALLIUM       = SaltVersion("Gallium"      , info=3021)
    GERMANIUM     = SaltVersion("Germanium"    , info=3022)
    ARSENIC       = SaltVersion("Arsenic"      , info=3023)
    SELENIUM      = SaltVersion("Selenium"     , info=3024)
    BROMINE       = SaltVersion("Bromine"      , info=3025)
    KRYPTON       = SaltVersion("Krypton"      , info=3026)
    RUBIDIUM      = SaltVersion("Rubidium"     , info=3027)
    STRONTIUM     = SaltVersion("Strontium"    , info=3028)
    YTTRIUM       = SaltVersion("Yttrium"      , info=3029)
    ZIRCONIUM     = SaltVersion("Zirconium"    , info=3030)
    NIOBIUM       = SaltVersion("Niobium"      , info=3031)
    MOLYBDENUM    = SaltVersion("Molybdenum"   , info=3032)
    TECHNETIUM    = SaltVersion("Technetium"   , info=3033)
    RUTHENIUM     = SaltVersion("Ruthenium"    , info=3034)
    RHODIUM       = SaltVersion("Rhodium"      , info=3035)
    PALLADIUM     = SaltVersion("Palladium"    , info=3036)
    SILVER        = SaltVersion("Silver"       , info=3037)
    CADMIUM       = SaltVersion("Cadmium"      , info=3038)
    INDIUM        = SaltVersion("Indium"       , info=3039)
    TIN           = SaltVersion("Tin"          , info=3040)
    ANTIMONY      = SaltVersion("Antimony"     , info=3041)
    TELLURIUM     = SaltVersion("Tellurium"    , info=3042)
    IODINE        = SaltVersion("Iodine"       , info=3043)
    XENON         = SaltVersion("Xenon"        , info=3044)
    CESIUM        = SaltVersion("Cesium"       , info=3045)
    BARIUM        = SaltVersion("Barium"       , info=3046)
    LANTHANUM     = SaltVersion("Lanthanum"    , info=3047)
    CERIUM        = SaltVersion("Cerium"       , info=3048)
    PRASEODYMIUM  = SaltVersion("Praseodymium" , info=3049)
    NEODYMIUM     = SaltVersion("Neodymium"    , info=3050)
    PROMETHIUM    = SaltVersion("Promethium"   , info=3051)
    SAMARIUM      = SaltVersion("Samarium"     , info=3052)
    EUROPIUM      = SaltVersion("Europium"     , info=3053)
    GADOLINIUM    = SaltVersion("Gadolinium"   , info=3054)
    TERBIUM       = SaltVersion("Terbium"      , info=3055)
    DYSPROSIUM    = SaltVersion("Dysprosium"   , info=3056)
    HOLMIUM       = SaltVersion("Holmium"      , info=3057)
    ERBIUM        = SaltVersion("Erbium"       , info=3058)
    THULIUM       = SaltVersion("Thulium"      , info=3059)
    YTTERBIUM     = SaltVersion("Ytterbium"    , info=3060)
    LUTETIUM      = SaltVersion("Lutetium"     , info=3061)
    HAFNIUM       = SaltVersion("Hafnium"      , info=3062)
    TANTALUM      = SaltVersion("Tantalum"     , info=3063)
    TUNGSTEN      = SaltVersion("Tungsten"     , info=3064)
    RHENIUM       = SaltVersion("Rhenium"      , info=3065)
    OSMIUM        = SaltVersion("Osmium"       , info=3066)
    IRIDIUM       = SaltVersion("Iridium"      , info=3067)
    PLATINUM      = SaltVersion("Platinum"     , info=3068)
    GOLD          = SaltVersion("Gold"         , info=3069)
    MERCURY       = SaltVersion("Mercury"      , info=3070)
    THALLIUM      = SaltVersion("Thallium"     , info=3071)
    LEAD          = SaltVersion("Lead"         , info=3072)
    BISMUTH       = SaltVersion("Bismuth"      , info=3073)
    POLONIUM      = SaltVersion("Polonium"     , info=3074)
    ASTATINE      = SaltVersion("Astatine"     , info=3075)
    RADON         = SaltVersion("Radon"        , info=3076)
    FRANCIUM      = SaltVersion("Francium"     , info=3077)
    RADIUM        = SaltVersion("Radium"       , info=3078)
    ACTINIUM      = SaltVersion("Actinium"     , info=3079)
    THORIUM       = SaltVersion("Thorium"      , info=3080)
    PROTACTINIUM  = SaltVersion("Protactinium" , info=3081)
    URANIUM       = SaltVersion("Uranium"      , info=3082)
    NEPTUNIUM     = SaltVersion("Neptunium"    , info=3083)
    PLUTONIUM     = SaltVersion("Plutonium"    , info=3084)
    AMERICIUM     = SaltVersion("Americium"    , info=3085)
    CURIUM        = SaltVersion("Curium"       , info=3086)
    BERKELIUM     = SaltVersion("Berkelium"    , info=3087)
    CALIFORNIUM   = SaltVersion("Californium"  , info=3088)
    EINSTEINIUM   = SaltVersion("Einsteinium"  , info=3089)
    FERMIUM       = SaltVersion("Fermium"      , info=3090)
    MENDELEVIUM   = SaltVersion("Mendelevium"  , info=3091)
    NOBELIUM      = SaltVersion("Nobelium"     , info=3092)
    LAWRENCIUM    = SaltVersion("Lawrencium"   , info=3093)
    RUTHERFORDIUM = SaltVersion("Rutherfordium", info=3094)
    DUBNIUM       = SaltVersion("Dubnium"      , info=3095)
    SEABORGIUM    = SaltVersion("Seaborgium"   , info=3096)
    BOHRIUM       = SaltVersion("Bohrium"      , info=3097)
    HASSIUM       = SaltVersion("Hassium"      , info=3098)
    MEITNERIUM    = SaltVersion("Meitnerium"   , info=3099)
    DARMSTADTIUM  = SaltVersion("Darmstadtium" , info=3100)
    ROENTGENIUM   = SaltVersion("Roentgenium"  , info=3101)
    COPERNICIUM   = SaltVersion("Copernicium"  , info=3102)
    NIHONIUM      = SaltVersion("Nihonium"     , info=3103)
    FLEROVIUM     = SaltVersion("Flerovium"    , info=3104)
    MOSCOVIUM     = SaltVersion("Moscovium"    , info=3105)
    LIVERMORIUM   = SaltVersion("Livermorium"  , info=3106)
    TENNESSINE    = SaltVersion("Tennessine"   , info=3107)
    OGANESSON     = SaltVersion("Oganesson"    , info=3108)
    # <---- Please refrain from fixing whitespace -----------------------------------
    # The idea is to keep this readable.
    # -------------------------------------------------------------------------------
    # pylint: enable=bad-whitespace
    # fmt: on

    @classmethod
    def versions(cls):
        if not cls._sorted_versions:
            cls._sorted_versions = sorted(
                (getattr(cls, name) for name in dir(cls) if name.isupper()),
                key=operator.attrgetter("info"),
            )
        return cls._sorted_versions

    @classmethod
    def current_release(cls):
        if cls._current_release is None:
            for version in cls.versions():
                if version.released is False:
                    cls._current_release = version
                    break
        return cls._current_release

    @classmethod
    def next_release(cls):
        if cls._next_release is None:
            next_release_ahead = False
            for version in cls.versions():
                if next_release_ahead:
                    cls._next_release = version
                    break
                if version == cls.current_release():
                    next_release_ahead = True
        return cls._next_release

    @classmethod
    def previous_release(cls):
        if cls._previous_release is None:
            previous = None
            for version in cls.versions():
                if version == cls.current_release():
                    break
                previous = version
            cls._previous_release = previous
        return cls._previous_release


class SaltStackVersion:
    """
    Handle SaltStack versions class.

    Knows how to parse ``git describe`` output, knows about release candidates
    and also supports version comparison.
    """

    __slots__ = (
        "name",
        "major",
        "minor",
        "bugfix",
        "mbugfix",
        "pre_type",
        "pre_num",
        "noc",
        "sha",
    )

    git_sha_regex = r"(?P<sha>g?[a-f0-9]{7,40})"

    git_describe_regex = re.compile(
        r"(?:[^\d]+)?(?P<major>[\d]{1,4})"
        r"(?:\.(?P<minor>[\d]{1,2}))?"
        r"(?:\.(?P<bugfix>[\d]{0,2}))?"
        r"(?:\.(?P<mbugfix>[\d]{0,2}))?"
        r"(?:(?P<pre_type>rc|a|b|alpha|beta|nb)(?P<pre_num>[\d]+))?"
        r"(?:(?:.*)(?:\+|-)(?P<noc>(?:0na|[\d]+|n/a))(?:-|\.)" + git_sha_regex + r")?"
    )
    git_sha_regex = r"^" + git_sha_regex

    git_sha_regex = re.compile(git_sha_regex)

    NAMES = {v.name: v.info for v in SaltVersionsInfo.versions()}
    LNAMES = {k.lower(): v for (k, v) in iter(NAMES.items())}
    VNAMES = {v: k for (k, v) in iter(NAMES.items())}
    RMATCH = {v[:2]: k for (k, v) in iter(NAMES.items())}

    def __init__(
        self,  # pylint: disable=C0103
        major,
        minor=None,
        bugfix=None,
        mbugfix=0,
        pre_type=None,
        pre_num=None,
        noc=0,
        sha=None,
    ):
        if isinstance(major, str):
            major = int(major)

        if isinstance(minor, str):
            if not minor:
                # Empty string
                minor = None
            else:
                minor = int(minor)
        if self.can_have_dot_zero(major):
            minor = minor if minor else 0

        if bugfix is None and not self.new_version(major=major):
            bugfix = 0
        elif isinstance(bugfix, str):
            if not bugfix:
                bugfix = None
            else:
                bugfix = int(bugfix)

        if mbugfix is None:
            mbugfix = 0
        elif isinstance(mbugfix, str):
            mbugfix = int(mbugfix)

        if pre_type is None:
            pre_type = ""
        if pre_num is None:
            pre_num = 0
        elif isinstance(pre_num, str):
            pre_num = int(pre_num)

        if noc is None:
            noc = 0
        elif isinstance(noc, str) and noc in ("0na", "n/a"):
            noc = -1
        elif isinstance(noc, str):
            noc = int(noc)

        self.major = major
        self.minor = minor
        self.bugfix = bugfix
        self.mbugfix = mbugfix
        self.pre_type = pre_type
        self.pre_num = pre_num
        if self.new_version(major):
            vnames_key = (major,)
        else:
            vnames_key = (major, minor)
        self.name = self.VNAMES.get(vnames_key)
        self.noc = noc
        self.sha = sha

    def new_version(self, major):
        """
        determine if using new versioning scheme
        """
        return bool(int(major) >= 3000 and int(major) < VERSION_LIMIT)

    def can_have_dot_zero(self, major):
        """
        determine if using new versioning scheme
        """
        return bool(int(major) >= 3006 and int(major) < VERSION_LIMIT)

    @classmethod
    def parse(cls, version_string):
        if version_string.lower() in cls.LNAMES:
            return cls.from_name(version_string)
        vstr = (
            version_string.decode()
            if isinstance(version_string, bytes)
            else version_string
        )
        match = cls.git_describe_regex.match(vstr)
        if not match:
            raise ValueError(f"Unable to parse version string: '{version_string}'")
        return cls(*match.groups())

    @classmethod
    def from_name(cls, name):
        if name.lower() not in cls.LNAMES:
            raise ValueError(f"Named version '{name}' is not known")
        return cls(*cls.LNAMES[name.lower()])

    @classmethod
    def from_last_named_version(cls):
        import salt.utils.versions

        salt.utils.versions.warn_until(
            SaltVersionsInfo.SULFUR,
            "The use of SaltStackVersion.from_last_named_version() is "
            "deprecated and set to be removed in {version}. Please use "
            "SaltStackVersion.current_release() instead.",
        )
        return cls.current_release()

    @classmethod
    def current_release(cls):
        return cls(*SaltVersionsInfo.current_release().info)

    @classmethod
    def next_release(cls):
        return cls(*SaltVersionsInfo.next_release().info)

    @property
    def sse(self):
        # Higher than 0.17, lower than first date based
        return 0 < self.major < 2014

    def min_info(self):
        info = [self.major]
        if self.new_version(self.major):
            if self.minor:
                info.append(self.minor)
            elif self.can_have_dot_zero(self.major):
                info.append(self.minor)
        else:
            info.extend([self.minor, self.bugfix, self.mbugfix])
        return info

    @property
    def info(self):
        return tuple(self.min_info())

    @property
    def pre_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num])
        return tuple(info)

    @property
    def noc_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num, self.noc])
        return tuple(info)

    @property
    def full_info(self):
        info = self.min_info()
        info.extend([self.pre_type, self.pre_num, self.noc, self.sha])
        return tuple(info)

    @property
    def full_info_all_versions(self):
        """
        Return the full info regardless
        of which versioning scheme we
        are using.
        """
        info = [
            self.major,
            self.minor,
            self.bugfix,
            self.mbugfix,
            self.pre_type,
            self.pre_num,
            self.noc,
            self.sha,
        ]
        return tuple(info)

    @property
    def string(self):
        if self.new_version(self.major):
            version_string = f"{self.major}"
            if self.minor:
                version_string = f"{self.major}.{self.minor}"
            if not self.minor and self.can_have_dot_zero(self.major):
                version_string = f"{self.major}.{self.minor}"
        else:
            version_string = f"{self.major}.{self.minor}.{self.bugfix}"
        if self.mbugfix:
            version_string += f".{self.mbugfix}"
        if self.pre_type:
            version_string += f"{self.pre_type}{self.pre_num}"
        if self.noc is not None and self.sha:
            noc = self.noc
            if noc < 0:
                noc = "0na"
            version_string += f"+{noc}.{self.sha}"
        return version_string

    @property
    def formatted_version(self):
        if self.name and self.major > 10000:
            version_string = self.name
            if self.sse:
                version_string += " Enterprise"
            version_string += " (Unreleased)"
            return version_string
        version_string = self.string
        if self.sse:
            version_string += " Enterprise"
        if self.new_version(self.major):
            rmatch_key = (self.major,)
        else:
            rmatch_key = (self.major, self.minor)
        if rmatch_key in self.RMATCH:
            version_string += f" ({self.RMATCH[rmatch_key]})"
        return version_string

    @property
    def pre_index(self):
        if self.new_version(self.major):
            pre_type = 2
            if not isinstance(self.minor, int):
                pre_type = 1
        else:
            pre_type = 4
        return pre_type

    def __str__(self):
        return self.string

    def __compare__(self, other, method):
        if not isinstance(other, SaltStackVersion):
            if isinstance(other, str):
                other = SaltStackVersion.parse(other)
            elif isinstance(other, (list, tuple)):
                other = SaltStackVersion(*other)
            else:
                raise ValueError(
                    f"Cannot instantiate Version from type '{type(other)}'"
                )
        pre_type = self.pre_index
        other_pre_type = other.pre_index
        other_noc_info = list(other.noc_info)
        noc_info = list(self.noc_info)

        if self.new_version(self.major):
            if self.minor and not other.minor:
                # We have minor information, the other side does not
                if self.minor > 0:
                    other_noc_info[1] = 0

            if not self.minor and other.minor:
                # The other side has minor information, we don't
                if other.minor > 0:
                    noc_info[1] = 0

        if self.pre_type and not other.pre_type:
            # We have pre-release information, the other side doesn't
            other_noc_info[other_pre_type] = "zzzzz"

        if not self.pre_type and other.pre_type:
            # The other side has pre-release information, we don't
            noc_info[pre_type] = "zzzzz"

        return method(tuple(noc_info), tuple(other_noc_info))

    def __lt__(self, other):
        return self.__compare__(other, lambda _self, _other: _self < _other)

    def __le__(self, other):
        return self.__compare__(other, lambda _self, _other: _self <= _other)

    def __eq__(self, other):
        return self.__compare__(other, lambda _self, _other: _self == _other)

    def __ne__(self, other):
        return self.__compare__(other, lambda _self, _other: _self != _other)

    def __ge__(self, other):
        return self.__compare__(other, lambda _self, _other: _self >= _other)

    def __gt__(self, other):
        return self.__compare__(other, lambda _self, _other: _self > _other)

    def __repr__(self):
        parts = []
        if self.name:
            parts.append(f"name='{self.name}'")
        parts.extend([f"major={self.major}", f"minor={self.minor}"])

        if self.new_version(self.major):
            if not self.can_have_dot_zero(self.major) and not self.minor:
                parts.remove("".join([x for x in parts if re.search("^minor*", x)]))
        else:
            parts.extend([f"bugfix={self.bugfix}"])

        if self.mbugfix:
            parts.append(f"minor-bugfix={self.mbugfix}")
        if self.pre_type:
            parts.append(f"{self.pre_type}={self.pre_num}")
        noc = self.noc
        if noc == -1:
            noc = "0na"
        if noc and self.sha:
            parts.extend([f"noc={noc}", f"sha={self.sha}"])
        return "<{} {}>".format(self.__class__.__name__, " ".join(parts))


# ----- Hardcoded Salt Codename Version Information ----------------------------------------------------------------->
#
#   There's no need to do anything here. The last released codename will be picked up
# --------------------------------------------------------------------------------------------------------------------
__saltstack_version__ = SaltStackVersion.current_release()
# <---- Hardcoded Salt Version Information ---------------------------------------------------------------------------


# ----- Dynamic/Runtime Salt Version Information -------------------------------------------------------------------->
def __discover_version(saltstack_version):
    # This might be a 'python setup.py develop' installation type. Let's
    # discover the version information at runtime.
    import subprocess

    if "SETUP_DIRNAME" in globals():
        # This is from the exec() call in Salt's setup.py
        cwd = SETUP_DIRNAME  # pylint: disable=E0602
        if not os.path.exists(os.path.join(cwd, ".git")):
            # This is not a Salt git checkout!!! Don't even try to parse...
            return saltstack_version
    else:
        cwd = os.path.abspath(os.path.dirname(__file__))
        if not os.path.exists(os.path.join(os.path.dirname(cwd), ".git")):
            # This is not a Salt git checkout!!! Don't even try to parse...
            return saltstack_version

    try:
        kwargs = dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd=cwd)

        if not sys.platform.startswith("win"):
            # Let's not import `salt.utils` for the above check
            kwargs["close_fds"] = True

        process = subprocess.Popen(
            [
                "git",
                "describe",
                "--tags",
                "--long",
                "--match",
                "v[0-9]*",
                "--always",
                "--candidates=150",
            ],
            **kwargs,
        )

        out, err = process.communicate()

        out = out.decode().strip()
        err = err.decode().strip()

        if not out or err:
            return saltstack_version

        if SaltStackVersion.git_sha_regex.match(out):
            # We only define the parsed SHA and set NOC as ??? (unknown)
            saltstack_version.sha = out.strip()
            saltstack_version.noc = -1
            return saltstack_version

        return SaltStackVersion.parse(out)

    except OSError as os_err:
        if os_err.errno != 2:
            # If the errno is not 2(The system cannot find the file
            # specified), raise the exception so it can be catch by the
            # developers
            raise
    return saltstack_version


def __get_version(saltstack_version):
    """
    If we can get a version provided at installation time or from Git, use
    that instead, otherwise we carry on.
    """
    _hardcoded_version_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "_version.txt"
    )
    if not os.path.exists(_hardcoded_version_file):
        return __discover_version(saltstack_version)
    with open(  # pylint: disable=resource-leakage
        _hardcoded_version_file, encoding="utf-8"
    ) as rfh:
        return SaltStackVersion.parse(rfh.read().strip())


# Get additional version information if available
__saltstack_version__ = __get_version(__saltstack_version__)
if __saltstack_version__.name:
    # Set SaltVersionsInfo._current_release to avoid lookups when finding previous and next releases
    SaltVersionsInfo._current_release = getattr(
        SaltVersionsInfo, __saltstack_version__.name.upper()
    )

# This function has executed once, we're done with it. Delete it!
del __get_version
# <---- Dynamic/Runtime Salt Version Information ---------------------------------------------------------------------


# ----- Common version related attributes - NO NEED TO CHANGE ------------------------------------------------------->
__version_info__ = __saltstack_version__.info
__version__ = __saltstack_version__.string
# <---- Common version related attributes - NO NEED TO CHANGE --------------------------------------------------------


def salt_information():
    """
    Report version of salt.
    """
    yield "Salt", __version__


def package_information():
    """
    Report package type
    """
    import salt.utils.package

    yield "Package Type", salt.utils.package.pkg_type()


def dependency_information(include_salt_cloud=False):
    """
    Report versions of library dependencies.
    """
    libs = [
        ("Jinja2", "jinja2", "__version__"),
        ("M2Crypto", "M2Crypto", "version"),
        ("msgpack", "msgpack", "version"),
        ("msgpack-pure", "msgpack_pure", "version"),
        ("pycrypto", "Crypto", "__version__"),
        ("pycryptodome", "Cryptodome", "version_info"),
        ("cryptography", "cryptography", "__version__"),
        ("PyYAML", "yaml", "__version__"),
        ("PyZMQ", "zmq", "__version__"),
        ("ZMQ", "zmq", "zmq_version"),
        ("Mako", "mako", "__version__"),
        ("Tornado", "tornado", "version"),
        ("timelib", "timelib", "version"),
        ("dateutil", "dateutil", "__version__"),
        ("pygit2", "pygit2", "__version__"),
        ("libgit2", "pygit2", "LIBGIT2_VERSION"),
        ("smmap", "smmap", "__version__"),
        ("cffi", "cffi", "__version__"),
        ("pycparser", "pycparser", "__version__"),
        ("gitdb", "gitdb", "__version__"),
        ("gitpython", "git", "__version__"),
        ("python-gnupg", "gnupg", "__version__"),
        ("mysql-python", "MySQLdb", "__version__"),
        ("cherrypy", "cherrypy", "__version__"),
        ("docker-py", "docker", "__version__"),
        ("packaging", "packaging", "__version__"),
        ("looseversion", "looseversion", None),
        ("relenv", "relenv", "__version__"),
    ]

    if include_salt_cloud:
        libs.append(
            ("Apache Libcloud", "libcloud", "__version__"),
        )

    def _sort_by_lowercased_name(entry):
        return entry[0].lower()

    for name, imp, attr in sorted(libs, key=_sort_by_lowercased_name):
        if imp is None:
            yield name, attr
            continue
        try:
            if attr is None:
                # Late import to reduce the needed available modules and libs
                # installed when running `python salt/version.py`
                from salt._compat import importlib_metadata

                version = importlib_metadata.version(imp)
                yield name, version
                continue
            imp = __import__(imp)
            version = getattr(imp, attr)
            if callable(version):
                version = version()
            if isinstance(version, (tuple, list)):
                version = ".".join(map(str, version))
            yield name, version
        except Exception:  # pylint: disable=broad-except
            yield name, None


def system_information():
    """
    Report system versions.
    """
    # Late import so that when getting called from setup.py does not break
    from salt.utils.platform import linux_distribution

    def system_version():
        """
        Return host system version.
        """

        lin_ver = linux_distribution()
        mac_ver = platform.mac_ver()
        win_ver = platform.win32_ver()

        # linux_distribution() will return a
        # distribution on OS X and Windows.
        # Check mac_ver and win_ver first,
        # then lin_ver.
        if mac_ver[0]:
            if isinstance(mac_ver[1], (tuple, list)) and "".join(mac_ver[1]):
                return " ".join([mac_ver[0], ".".join(mac_ver[1]), mac_ver[2]])
            else:
                return " ".join([mac_ver[0], mac_ver[2]])
        elif win_ver[0]:
            return " ".join(win_ver)
        elif lin_ver[0]:
            return " ".join(lin_ver)
        else:
            return ""

    if platform.win32_ver()[0]:
        # Get the version and release info based on the Windows Operating
        # System Product Name. As long as Microsoft maintains a similar format
        # this should be future proof
        import win32api  # pylint: disable=3rd-party-module-not-gated
        import win32con  # pylint: disable=3rd-party-module-not-gated

        # Get the product name from the registry
        hkey = win32con.HKEY_LOCAL_MACHINE
        key = "SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion"
        value_name = "ProductName"
        reg_handle = win32api.RegOpenKey(hkey, key)

        # Returns a tuple of (product_name, value_type)
        product_name, _ = win32api.RegQueryValueEx(reg_handle, value_name)

        version = "Unknown"
        release = ""
        if "Server" in product_name:
            for item in product_name.split(" "):
                # If it's all digits, then it's version
                if re.match(r"\d+", item):
                    version = item
                # If it starts with R and then numbers, it's the release
                # ie: R2
                if re.match(r"^R\d+$", item):
                    release = item
            release = f"{version}Server{release}"
        else:
            for item in product_name.split(" "):
                # If it's a number, decimal number, Thin or Vista, then it's the
                # version
                if re.match(r"^(\d+(\.\d+)?)|Thin|Vista$", item):
                    version = item
            release = version

        _, ver, service_pack, extra = platform.win32_ver()
        version = " ".join([release, ver, service_pack, extra])
    else:
        version = system_version()
        release = platform.release()

    system = [
        ("system", platform.system()),
        ("dist", " ".join(linux_distribution(full_distribution_name=False))),
        ("release", release),
        ("machine", platform.machine()),
        ("version", version),
        ("locale", __salt_system_encoding__),
    ]

    for name, attr in system:
        yield name, attr
        continue


def extensions_information():
    """
    Gather infomation about any installed salt extensions
    """
    # Late import
    import salt.utils.entrypoints

    extensions = {}
    for entry_point in salt.utils.entrypoints.iter_entry_points("salt.loader"):
        dist_nv = salt.utils.entrypoints.name_and_version_from_entry_point(entry_point)
        if not dist_nv:
            continue
        if dist_nv.name in extensions:
            continue
        extensions[dist_nv.name] = dist_nv.version
    return extensions


def versions_information(include_salt_cloud=False, include_extensions=True):
    """
    Report the versions of dependent software.
    """
    py_info = [
        ("Python", sys.version.rsplit("\n", maxsplit=1)[0].strip()),
    ]
    salt_info = list(salt_information())
    lib_info = list(dependency_information(include_salt_cloud))
    sys_info = list(system_information())
    package_info = list(package_information())

    info = {
        "Salt Version": dict(salt_info),
        "Python Version": dict(py_info),
        "Dependency Versions": dict(lib_info),
        "System Versions": dict(sys_info),
        "Salt Package Information": dict(package_info),
    }
    if include_extensions:
        extensions_info = extensions_information()
        if extensions_info:
            info["Salt Extensions"] = extensions_info
    return info


def versions_report(include_salt_cloud=False, include_extensions=True):
    """
    Yield each version properly formatted for console output.
    """
    ver_info = versions_information(
        include_salt_cloud=include_salt_cloud, include_extensions=include_extensions
    )
    not_installed = "Not Installed"
    ns_pad = len(not_installed)
    lib_pad = max(len(name) for name in ver_info["Dependency Versions"])
    sys_pad = max(len(name) for name in ver_info["System Versions"])
    if include_extensions and "Salt Extensions" in ver_info:
        ext_pad = max(len(name) for name in ver_info["Salt Extensions"])
    else:
        ext_pad = 1
    padding = max(lib_pad, sys_pad, ns_pad, ext_pad) + 1

    fmt = "{0:>{pad}}: {1}"
    info = []
    for ver_type in (
        "Salt Version",
        "Python Version",
        "Dependency Versions",
        "Salt Extensions",
        "Salt Package Information",
        "System Versions",
    ):
        if ver_type == "Salt Extensions" and ver_type not in ver_info:
            # No salt Extensions to report
            continue
        info.append(f"{ver_type}:")
        # List dependencies in alphabetical, case insensitive order
        for name in sorted(ver_info[ver_type], key=lambda x: x.lower()):
            ver = fmt.format(
                name, ver_info[ver_type][name] or not_installed, pad=padding
            )
            info.append(ver)
        info.append(" ")

    yield from info


def _parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--next-release", help="Return the next release", action="store_true"
    )
    parser.add_argument("--parse", help="Parse the passed string as a salt version")
    # When pip installing we pass in other args to this script.
    # This allows us to catch those args but not use them
    parser.add_argument("unknown", nargs=argparse.REMAINDER)
    return parser.parse_args()


if __name__ == "__main__":
    args = _parser()
    if args.next_release:
        print(__saltstack_version__.next_release())
    elif args.parse:
        try:
            print(SaltStackVersion.parse(args.parse))
        except Exception as exc:  # pylint: disable=broad-except
            print(f"Failed to parse '{args.parse}' as a salt version: {exc}")
            sys.exit(1)
    else:
        print(__version__)



"""
Neutron class
"""

import logging

import salt.utils.versions
from salt import exceptions

# pylint: disable=import-error
HAS_NEUTRON = False
try:
    from neutronclient.shell import NeutronShell
    from neutronclient.v2_0 import client

    HAS_NEUTRON = True
except ImportError:
    pass

HAS_KEYSTONEAUTH = False
try:
    import keystoneauth1.loading
    import keystoneauth1.session

    HAS_KEYSTONEAUTH = True
except ImportError:
    pass
# pylint: enable=import-error


# Get logging started
log = logging.getLogger(__name__)


def check_neutron():
    return HAS_NEUTRON


def check_keystone():
    return HAS_KEYSTONEAUTH


def sanitize_neutronclient(kwargs):
    variables = (
        "username",
        "user_id",
        "password",
        "token",
        "tenant_name",
        "tenant_id",
        "auth_url",
        "service_type",
        "endpoint_type",
        "region_name",
        "verify",
        "endpoint_url",
        "timeout",
        "insecure",
        "ca_cert",
        "retries",
        "raise_error",
        "session",
        "auth",
    )
    ret = {}
    for var in kwargs.keys():
        if var in variables:
            ret[var] = kwargs[var]

    return ret


# Function alias to not shadow built-ins
class SaltNeutron(NeutronShell):
    """
    Class for all neutronclient functions
    """

    def __init__(
        self,
        username,
        tenant_name,
        auth_url,
        password=None,
        region_name=None,
        service_type="network",
        os_auth_plugin=None,
        use_keystoneauth=False,
        **kwargs
    ):
        """
        Set up neutron credentials
        """
        salt.utils.versions.warn_until(
            3008,
            "The neutron module has been deprecated and will be removed in {version}.\n"
            "This includes\n"
            "* salt.utils.openstack.neutron\n"
            "* salt.modules.neutron\n"
            "* salt.pillar.neutron\n"
            "Please migrate to neutronng.\n"
            "salt.modules.neutron -> salt.modules.neutronng\n"
            "salt.pillar.neutron -> salt.pillar.neutronng\n"
            "Please update to using the neutronng module",
        )
        if not HAS_NEUTRON:
            return None

        elif all([use_keystoneauth, HAS_KEYSTONEAUTH]):
            self._new_init(
                username=username,
                project_name=tenant_name,
                auth_url=auth_url,
                region_name=region_name,
                service_type=service_type,
                os_auth_plugin=os_auth_plugin,
                password=password,
                **kwargs
            )
        else:
            self._old_init(
                username=username,
                tenant_name=tenant_name,
                auth_url=auth_url,
                region_name=region_name,
                service_type=service_type,
                os_auth_plugin=os_auth_plugin,
                password=password,
                **kwargs
            )

    def _new_init(
        self,
        username,
        project_name,
        auth_url,
        region_name,
        service_type,
        password,
        os_auth_plugin,
        auth=None,
        verify=True,
        **kwargs
    ):
        if auth is None:
            auth = {}

        loader = keystoneauth1.loading.get_plugin_loader(os_auth_plugin or "password")

        self.client_kwargs = kwargs.copy()
        self.kwargs = auth.copy()

        self.kwargs["username"] = username
        self.kwargs["project_name"] = project_name
        self.kwargs["auth_url"] = auth_url
        self.kwargs["password"] = password
        if auth_url.endswith("3"):
            self.kwargs["user_domain_name"] = kwargs.get("user_domain_name", "default")
            self.kwargs["project_domain_name"] = kwargs.get(
                "project_domain_name", "default"
            )

        self.client_kwargs["region_name"] = region_name
        self.client_kwargs["service_type"] = service_type

        self.client_kwargs = sanitize_neutronclient(self.client_kwargs)
        options = loader.load_from_options(**self.kwargs)
        self.session = keystoneauth1.session.Session(auth=options, verify=verify)
        self.network_conn = client.Client(session=self.session, **self.client_kwargs)

    def _old_init(
        self,
        username,
        tenant_name,
        auth_url,
        region_name,
        service_type,
        password,
        os_auth_plugin,
        auth=None,
        verify=True,
        **kwargs
    ):
        self.kwargs = kwargs.copy()

        self.kwargs["username"] = username
        self.kwargs["tenant_name"] = tenant_name
        self.kwargs["auth_url"] = auth_url
        self.kwargs["service_type"] = service_type
        self.kwargs["password"] = password
        self.kwargs["region_name"] = region_name
        self.kwargs["verify"] = verify

        self.kwargs = sanitize_neutronclient(self.kwargs)

        self.network_conn = client.Client(**self.kwargs)

    @staticmethod
    def _fetch(resources, name_or_id):
        ret = []
        for resource in resources:
            if resource["id"] == name_or_id:
                return resource
            if resource.get("name") == name_or_id:
                ret.append(resource)
        if not ret:
            raise exceptions.MinionError("Resource not found.")
        elif len(ret) >= 2:
            raise exceptions.MinionError("Multiple resource matches found.")
        else:
            return ret[0]

    def _find_port_id(self, resource):
        resource = self._fetch_port(resource)
        return resource["id"]

    def _find_network_id(self, resource):
        resource = self._fetch_network(resource)
        return resource["id"]

    def _find_subnet_id(self, resource):
        resource = self._fetch_subnet(resource)
        return resource["id"]

    def _find_router_id(self, resource):
        resource = self._fetch_router(resource)
        return resource["id"]

    def _find_security_group_id(self, resource):
        resource = self._fetch_security_group(resource)
        return resource["id"]

    def _find_vpnservice_id(self, resource):
        resource = self._fetch_vpnservice(resource)
        return resource["id"]

    def _find_ipsec_site_connection_id(self, resource):
        resource = self._fetch_ipsec_site_connection(resource)
        return resource["id"]

    def _find_ikepolicy_id(self, resource):
        resource = self._fetch_ikepolicy(resource)
        return resource["id"]

    def _find_ipsecpolicy_id(self, resource):
        resource = self._fetch_ipsecpolicy(resource)
        return resource["id"]

    def _find_firewall_rule_id(self, resource):
        resource = self._fetch_firewall_rule(resource)
        return resource["id"]

    def _fetch_port(self, name_or_id):
        resources = self.list_ports()["ports"]
        return self._fetch(resources, name_or_id)

    def _fetch_network(self, name_or_id):
        resources = self.list_networks()["networks"]
        return self._fetch(resources, name_or_id)

    def _fetch_subnet(self, name_or_id):
        resources = self.list_subnets()["subnets"]
        return self._fetch(resources, name_or_id)

    def _fetch_router(self, name_or_id):
        resources = self.list_routers()["routers"]
        return self._fetch(resources, name_or_id)

    def _fetch_security_group(self, name_or_id):
        resources = self.list_security_groups()["security_groups"]
        return self._fetch(resources, name_or_id)

    def _fetch_vpnservice(self, name_or_id):
        resources = self.list_vpnservices()["vpnservices"]
        return self._fetch(resources, name_or_id)

    def _fetch_ipsec_site_connection(self, name_or_id):
        resources = self.list_ipsec_site_connections()["ipsec_site_connections"]
        return self._fetch(resources, name_or_id)

    def _fetch_ikepolicy(self, name_or_id):
        resources = self.list_ikepolicies()["ikepolicies"]
        return self._fetch(resources, name_or_id)

    def _fetch_ipsecpolicy(self, name_or_id):
        resources = self.list_ipsecpolicies()["ipsecpolicies"]
        return self._fetch(resources, name_or_id)

    def _fetch_firewall_rule(self, name_or_id):
        resources = self.list_firewall_rules()["firewall_rules"]
        return self._fetch(resources, name_or_id)

    def _fetch_firewall(self, name_or_id):
        resources = self.list_firewalls()["firewalls"]
        return self._fetch(resources, name_or_id)

    def get_quotas_tenant(self):
        """
        Fetches tenant info in server's context
        for following quota operation
        """
        return self.get_quotas_tenant()

    def list_quotas(self):
        """
        Fetches all tenants quotas
        """
        return self.network_conn.list_quotas()

    def show_quota(self, tenant_id):
        """
        Fetches information of a certain tenant's quotas
        """
        return self.network_conn.show_quota(tenant_id=tenant_id)

    def update_quota(
        self,
        tenant_id,
        subnet=None,
        router=None,
        network=None,
        floatingip=None,
        port=None,
        sec_grp=None,
        sec_grp_rule=None,
    ):
        """
        Update a tenant's quota
        """
        body = {}
        if subnet:
            body["subnet"] = subnet
        if router:
            body["router"] = router
        if network:
            body["network"] = network
        if floatingip:
            body["floatingip"] = floatingip
        if port:
            body["port"] = port
        if sec_grp:
            body["security_group"] = sec_grp
        if sec_grp_rule:
            body["security_group_rule"] = sec_grp_rule
        return self.network_conn.update_quota(tenant_id=tenant_id, body={"quota": body})

    def delete_quota(self, tenant_id):
        """
        Delete the specified tenant's quota value
        """
        ret = self.network_conn.delete_quota(tenant_id=tenant_id)
        return ret if ret else True

    def list_extensions(self):
        """
        Fetches a list of all extensions on server side
        """
        return self.network_conn.list_extensions()

    def list_ports(self):
        """
        Fetches a list of all ports for a tenant
        """
        return self.network_conn.list_ports()

    def show_port(self, port):
        """
        Fetches information of a certain port
        """
        return self._fetch_port(port)

    def create_port(self, name, network, device_id=None, admin_state_up=True):
        """
        Creates a new port
        """
        net_id = self._find_network_id(network)
        body = {"admin_state_up": admin_state_up, "name": name, "network_id": net_id}
        if device_id:
            body["device_id"] = device_id
        return self.network_conn.create_port(body={"port": body})

    def update_port(self, port, name, admin_state_up=True):
        """
        Updates a port
        """
        port_id = self._find_port_id(port)
        body = {"name": name, "admin_state_up": admin_state_up}
        return self.network_conn.update_port(port=port_id, body={"port": body})

    def delete_port(self, port):
        """
        Deletes the specified port
        """
        port_id = self._find_port_id(port)
        ret = self.network_conn.delete_port(port=port_id)
        return ret if ret else True

    def list_networks(self):
        """
        Fetches a list of all networks for a tenant
        """
        return self.network_conn.list_networks()

    def show_network(self, network):
        """
        Fetches information of a certain network
        """
        return self._fetch_network(network)

    def create_network(
        self,
        name,
        admin_state_up=True,
        router_ext=None,
        network_type=None,
        physical_network=None,
        segmentation_id=None,
        shared=None,
        vlan_transparent=None,
    ):
        """
        Creates a new network
        """
        body = {"name": name, "admin_state_up": admin_state_up}
        if router_ext:
            body["router:external"] = router_ext
        if network_type:
            body["provider:network_type"] = network_type
        if physical_network:
            body["provider:physical_network"] = physical_network
        if segmentation_id:
            body["provider:segmentation_id"] = segmentation_id
        if shared:
            body["shared"] = shared
        if vlan_transparent:
            body["vlan_transparent"] = vlan_transparent
        return self.network_conn.create_network(body={"network": body})

    def update_network(self, network, name):
        """
        Updates a network
        """
        net_id = self._find_network_id(network)
        return self.network_conn.update_network(
            network=net_id, body={"network": {"name": name}}
        )

    def delete_network(self, network):
        """
        Deletes the specified network
        """
        net_id = self._find_network_id(network)
        ret = self.network_conn.delete_network(network=net_id)
        return ret if ret else True

    def list_subnets(self):
        """
        Fetches a list of all networks for a tenant
        """
        return self.network_conn.list_subnets()

    def show_subnet(self, subnet):
        """
        Fetches information of a certain subnet
        """
        return self._fetch_subnet(subnet)

    def create_subnet(self, network, cidr, name=None, ip_version=4):
        """
        Creates a new subnet
        """
        net_id = self._find_network_id(network)
        body = {
            "cidr": cidr,
            "ip_version": ip_version,
            "network_id": net_id,
            "name": name,
        }
        return self.network_conn.create_subnet(body={"subnet": body})

    def update_subnet(self, subnet, name=None):
        """
        Updates a subnet
        """
        subnet_id = self._find_subnet_id(subnet)
        return self.network_conn.update_subnet(
            subnet=subnet_id, body={"subnet": {"name": name}}
        )

    def delete_subnet(self, subnet):
        """
        Deletes the specified subnet
        """
        subnet_id = self._find_subnet_id(subnet)
        ret = self.network_conn.delete_subnet(subnet=subnet_id)
        return ret if ret else True

    def list_routers(self):
        """
        Fetches a list of all routers for a tenant
        """
        return self.network_conn.list_routers()

    def show_router(self, router):
        """
        Fetches information of a certain router
        """
        return self._fetch_router(router)

    def create_router(self, name, ext_network=None, admin_state_up=True):
        """
        Creates a new router
        """
        body = {"name": name, "admin_state_up": admin_state_up}
        if ext_network:
            net_id = self._find_network_id(ext_network)
            body["external_gateway_info"] = {"network_id": net_id}
        return self.network_conn.create_router(body={"router": body})

    def update_router(self, router, name=None, admin_state_up=None, **kwargs):
        """
        Updates a router
        """
        router_id = self._find_router_id(router)
        body = {}

        if "ext_network" in kwargs:
            if kwargs.get("ext_network") is None:
                body["external_gateway_info"] = None
            else:
                net_id = self._find_network_id(kwargs.get("ext_network"))
                body["external_gateway_info"] = {"network_id": net_id}
        if name is not None:
            body["name"] = name
        if admin_state_up is not None:
            body["admin_state_up"] = admin_state_up
        return self.network_conn.update_router(router=router_id, body={"router": body})

    def delete_router(self, router):
        """
        Delete the specified router
        """
        router_id = self._find_router_id(router)
        ret = self.network_conn.delete_router(router=router_id)
        return ret if ret else True

    def add_interface_router(self, router, subnet):
        """
        Adds an internal network interface to the specified router
        """
        router_id = self._find_router_id(router)
        subnet_id = self._find_subnet_id(subnet)
        return self.network_conn.add_interface_router(
            router=router_id, body={"subnet_id": subnet_id}
        )

    def remove_interface_router(self, router, subnet):
        """
        Removes an internal network interface from the specified router
        """
        router_id = self._find_router_id(router)
        subnet_id = self._find_subnet_id(subnet)
        return self.network_conn.remove_interface_router(
            router=router_id, body={"subnet_id": subnet_id}
        )

    def add_gateway_router(self, router, network):
        """
        Adds an external network gateway to the specified router
        """
        router_id = self._find_router_id(router)
        net_id = self._find_network_id(network)
        return self.network_conn.add_gateway_router(
            router=router_id, body={"network_id": net_id}
        )

    def remove_gateway_router(self, router):
        """
        Removes an external network gateway from the specified router
        """
        router_id = self._find_router_id(router)
        return self.network_conn.remove_gateway_router(router=router_id)

    def list_floatingips(self):
        """
        Fetch a list of all floatingips for a tenant
        """
        return self.network_conn.list_floatingips()

    def show_floatingip(self, floatingip_id):
        """
        Fetches information of a certain floatingip
        """
        return self.network_conn.show_floatingip(floatingip_id)

    def create_floatingip(self, floating_network, port=None):
        """
        Creates a new floatingip
        """
        net_id = self._find_network_id(floating_network)
        body = {"floating_network_id": net_id}
        if port:
            port_id = self._find_port_id(port)
            body["port_id"] = port_id

        return self.network_conn.create_floatingip(body={"floatingip": body})

    def update_floatingip(self, floatingip_id, port=None):
        """
        Updates a floatingip, disassociates the floating ip if
        port is set to `None`
        """
        if port is None:
            body = {"floatingip": {}}
        else:
            port_id = self._find_port_id(port)
            body = {"floatingip": {"port_id": port_id}}
        return self.network_conn.update_floatingip(floatingip=floatingip_id, body=body)

    def delete_floatingip(self, floatingip_id):
        """
        Deletes the specified floatingip
        """
        ret = self.network_conn.delete_floatingip(floatingip_id)
        return ret if ret else True

    def list_security_groups(self):
        """
        Fetches a list of all security groups for a tenant
        """
        return self.network_conn.list_security_groups()

    def show_security_group(self, sec_grp):
        """
        Fetches information of a certain security group
        """
        return self._fetch_security_group(sec_grp)

    def create_security_group(self, name, desc=None):
        """
        Creates a new security group
        """
        body = {"security_group": {"name": name, "description": desc}}
        return self.network_conn.create_security_group(body=body)

    def update_security_group(self, sec_grp, name=None, desc=None):
        """
        Updates a security group
        """
        sec_grp_id = self._find_security_group_id(sec_grp)
        body = {"security_group": {}}
        if name:
            body["security_group"]["name"] = name
        if desc:
            body["security_group"]["description"] = desc
        return self.network_conn.update_security_group(sec_grp_id, body=body)

    def delete_security_group(self, sec_grp):
        """
        Deletes the specified security group
        """
        sec_grp_id = self._find_security_group_id(sec_grp)
        ret = self.network_conn.delete_security_group(sec_grp_id)
        return ret if ret else True

    def list_security_group_rules(self):
        """
        Fetches a list of all security group rules for a tenant
        """
        return self.network_conn.list_security_group_rules()

    def show_security_group_rule(self, sec_grp_rule_id):
        """
        Fetches information of a certain security group rule
        """
        return self.network_conn.show_security_group_rule(sec_grp_rule_id)[
            "security_group_rule"
        ]

    def create_security_group_rule(
        self,
        sec_grp,
        remote_grp_id=None,
        direction="ingress",
        protocol=None,
        port_range_min=None,
        port_range_max=None,
        ether=None,
    ):
        """
        Creates a new security group rule
        """
        sec_grp_id = self._find_security_group_id(sec_grp)
        body = {
            "security_group_id": sec_grp_id,
            "remote_group_id": remote_grp_id,
            "direction": direction,
            "protocol": protocol,
            "port_range_min": port_range_min,
            "port_range_max": port_range_max,
            "ethertype": ether,
        }
        return self.network_conn.create_security_group_rule(
            body={"security_group_rule": body}
        )

    def delete_security_group_rule(self, sec_grp_rule_id):
        """
        Deletes the specified security group rule
        """
        ret = self.network_conn.delete_security_group_rule(
            security_group_rule=sec_grp_rule_id
        )
        return ret if ret else True

    def list_vpnservices(self, retrieve_all=True, **kwargs):
        """
        Fetches a list of all configured VPN services for a tenant
        """
        return self.network_conn.list_vpnservices(retrieve_all, **kwargs)

    def show_vpnservice(self, vpnservice, **kwargs):
        """
        Fetches information of a specific VPN service
        """
        vpnservice_id = self._find_vpnservice_id(vpnservice)
        return self.network_conn.show_vpnservice(vpnservice_id, **kwargs)

    def create_vpnservice(self, subnet, router, name, admin_state_up=True):
        """
        Creates a new VPN service
        """
        subnet_id = self._find_subnet_id(subnet)
        router_id = self._find_router_id(router)
        body = {
            "subnet_id": subnet_id,
            "router_id": router_id,
            "name": name,
            "admin_state_up": admin_state_up,
        }
        return self.network_conn.create_vpnservice(body={"vpnservice": body})

    def update_vpnservice(self, vpnservice, desc):
        """
        Updates a VPN service
        """
        vpnservice_id = self._find_vpnservice_id(vpnservice)
        body = {"description": desc}
        return self.network_conn.update_vpnservice(
            vpnservice_id, body={"vpnservice": body}
        )

    def delete_vpnservice(self, vpnservice):
        """
        Deletes the specified VPN service
        """
        vpnservice_id = self._find_vpnservice_id(vpnservice)
        ret = self.network_conn.delete_vpnservice(vpnservice_id)
        return ret if ret else True

    def list_ipsec_site_connections(self):
        """
        Fetches all configured IPsec Site Connections for a tenant
        """
        return self.network_conn.list_ipsec_site_connections()

    def show_ipsec_site_connection(self, ipsec_site_connection):
        """
        Fetches information of a specific IPsecSiteConnection
        """
        return self._fetch_ipsec_site_connection(ipsec_site_connection)

    def create_ipsec_site_connection(
        self,
        name,
        ipsecpolicy,
        ikepolicy,
        vpnservice,
        peer_cidrs,
        peer_address,
        peer_id,
        psk,
        admin_state_up=True,
        **kwargs
    ):
        """
        Creates a new IPsecSiteConnection
        """
        ipsecpolicy_id = self._find_ipsecpolicy_id(ipsecpolicy)
        ikepolicy_id = self._find_ikepolicy_id(ikepolicy)
        vpnservice_id = self._find_vpnservice_id(vpnservice)
        body = {
            "psk": psk,
            "ipsecpolicy_id": ipsecpolicy_id,
            "admin_state_up": admin_state_up,
            "peer_cidrs": [peer_cidrs],
            "ikepolicy_id": ikepolicy_id,
            "vpnservice_id": vpnservice_id,
            "peer_address": peer_address,
            "peer_id": peer_id,
            "name": name,
        }
        if "initiator" in kwargs:
            body["initiator"] = kwargs["initiator"]
        if "mtu" in kwargs:
            body["mtu"] = kwargs["mtu"]
        if "dpd_action" in kwargs:
            body["dpd"] = {"action": kwargs["dpd_action"]}
        if "dpd_interval" in kwargs:
            if "dpd" not in body:
                body["dpd"] = {}
            body["dpd"]["interval"] = kwargs["dpd_interval"]
        if "dpd_timeout" in kwargs:
            if "dpd" not in body:
                body["dpd"] = {}
            body["dpd"]["timeout"] = kwargs["dpd_timeout"]
        return self.network_conn.create_ipsec_site_connection(
            body={"ipsec_site_connection": body}
        )

    def delete_ipsec_site_connection(self, ipsec_site_connection):
        """
        Deletes the specified IPsecSiteConnection
        """
        ipsec_site_connection_id = self._find_ipsec_site_connection_id(
            ipsec_site_connection
        )
        ret = self.network_conn.delete_ipsec_site_connection(ipsec_site_connection_id)
        return ret if ret else True

    def list_ikepolicies(self):
        """
        Fetches a list of all configured IKEPolicies for a tenant
        """
        return self.network_conn.list_ikepolicies()

    def show_ikepolicy(self, ikepolicy):
        """
        Fetches information of a specific IKEPolicy
        """
        return self._fetch_ikepolicy(ikepolicy)

    def create_ikepolicy(self, name, **kwargs):
        """
        Creates a new IKEPolicy
        """
        body = {"name": name}
        if "phase1_negotiation_mode" in kwargs:
            body["phase1_negotiation_mode"] = kwargs["phase1_negotiation_mode"]
        if "auth_algorithm" in kwargs:
            body["auth_algorithm"] = kwargs["auth_algorithm"]
        if "encryption_algorithm" in kwargs:
            body["encryption_algorithm"] = kwargs["encryption_algorithm"]
        if "pfs" in kwargs:
            body["pfs"] = kwargs["pfs"]
        if "ike_version" in kwargs:
            body["ike_version"] = kwargs["ike_version"]
        if "units" in kwargs:
            body["lifetime"] = {"units": kwargs["units"]}
        if "value" in kwargs:
            if "lifetime" not in body:
                body["lifetime"] = {}
            body["lifetime"]["value"] = kwargs["value"]
        return self.network_conn.create_ikepolicy(body={"ikepolicy": body})

    def delete_ikepolicy(self, ikepolicy):
        """
        Deletes the specified IKEPolicy
        """
        ikepolicy_id = self._find_ikepolicy_id(ikepolicy)
        ret = self.network_conn.delete_ikepolicy(ikepolicy_id)
        return ret if ret else True

    def list_ipsecpolicies(self):
        """
        Fetches a list of all configured IPsecPolicies for a tenant
        """
        return self.network_conn.list_ipsecpolicies()

    def show_ipsecpolicy(self, ipsecpolicy):
        """
        Fetches information of a specific IPsecPolicy
        """
        return self._fetch_ipsecpolicy(ipsecpolicy)

    def create_ipsecpolicy(self, name, **kwargs):
        """
        Creates a new IPsecPolicy
        """
        body = {"name": name}
        if "transform_protocol" in kwargs:
            body["transform_protocol"] = kwargs["transform_protocol"]
        if "auth_algorithm" in kwargs:
            body["auth_algorithm"] = kwargs["auth_algorithm"]
        if "encapsulation_mode" in kwargs:
            body["encapsulation_mode"] = kwargs["encapsulation_mode"]
        if "encryption_algorithm" in kwargs:
            body["encryption_algorithm"] = kwargs["encryption_algorithm"]
        if "pfs" in kwargs:
            body["pfs"] = kwargs["pfs"]
        if "units" in kwargs:
            body["lifetime"] = {"units": kwargs["units"]}
        if "value" in kwargs:
            if "lifetime" not in body:
                body["lifetime"] = {}
            body["lifetime"]["value"] = kwargs["value"]
        return self.network_conn.create_ipsecpolicy(body={"ipsecpolicy": body})

    def delete_ipsecpolicy(self, ipseecpolicy):
        """
        Deletes the specified IPsecPolicy
        """
        ipseecpolicy_id = self._find_ipsecpolicy_id(ipseecpolicy)
        ret = self.network_conn.delete_ipsecpolicy(ipseecpolicy_id)
        return ret if ret else True

    def list_firewall_rules(self):
        """
        Fetches a list of all configured firewall rules for a tenant
        """
        return self.network_conn.list_firewall_rules()

    def show_firewall_rule(self, firewall_rule):
        """
        Fetches information of a specific firewall rule
        """
        return self._fetch_firewall_rule(firewall_rule)

    def create_firewall_rule(self, protocol, action, **kwargs):
        """
        Create a new firlwall rule
        """
        body = {"protocol": protocol, "action": action}
        if "tenant_id" in kwargs:
            body["tenant_id"] = kwargs["tenant_id"]
        if "name" in kwargs:
            body["name"] = kwargs["name"]
        if "description" in kwargs:
            body["description"] = kwargs["description"]
        if "ip_version" in kwargs:
            body["ip_version"] = kwargs["ip_version"]
        if "source_ip_address" in kwargs:
            body["source_ip_address"] = kwargs["source_ip_address"]
        if "destination_port" in kwargs:
            body["destination_port"] = kwargs["destination_port"]
        if "shared" in kwargs:
            body["shared"] = kwargs["shared"]
        if "enabled" in kwargs:
            body["enabled"] = kwargs["enabled"]
        return self.network_conn.create_firewall_rule(body={"firewall_rule": body})

    def delete_firewall_rule(self, firewall_rule):
        """
        Deletes the specified firewall rule
        """
        firewall_rule_id = self._find_firewall_rule_id(firewall_rule)
        ret = self.network_conn.delete_firewall_rule(firewall_rule_id)
        return ret if ret else True

    def update_firewall_rule(
        self,
        firewall_rule,
        protocol=None,
        action=None,
        name=None,
        description=None,
        ip_version=None,
        source_ip_address=None,
        destination_ip_address=None,
        source_port=None,
        destination_port=None,
        shared=None,
        enabled=None,
    ):
        """
        Update a firewall rule
        """
        body = {}
        if protocol:
            body["protocol"] = protocol
        if action:
            body["action"] = action
        if name:
            body["name"] = name
        if description:
            body["description"] = description
        if ip_version:
            body["ip_version"] = ip_version
        if source_ip_address:
            body["source_ip_address"] = source_ip_address
        if destination_ip_address:
            body["destination_ip_address"] = destination_ip_address
        if source_port:
            body["source_port"] = source_port
        if destination_port:
            body["destination_port"] = destination_port
        if shared:
            body["shared"] = shared
        if enabled:
            body["enabled"] = enabled
        return self.network_conn.update_firewall_rule(
            firewall_rule, body={"firewall_rule": body}
        )

    def list_firewalls(self):
        """
        Fetches a list of all firewalls for a tenant
        """
        return self.network_conn.list_firewalls()

    def show_firewall(self, firewall):
        """
        Fetches information of a specific firewall
        """
        return self._fetch_firewall(firewall)

    def list_l3_agent_hosting_routers(self, router):
        """
        List L3 agents.
        """
        return self.network_conn.list_l3_agent_hosting_routers(router)

    def list_agents(self):
        """
        List agents.
        """
        return self.network_conn.list_agents()


# The following is a list of functions that need to be incorporated in the
# neutron module. This list should be updated as functions are added.
#
# update_ipsec_site_connection
#                       Updates an IPsecSiteConnection.
# update_ikepolicy      Updates an IKEPolicy
# update_ipsecpolicy    Updates an IPsecPolicy
# list_vips             Fetches a list of all load balancer vips for a tenant.
# show_vip              Fetches information of a certain load balancer vip.
# create_vip            Creates a new load balancer vip.
# update_vip            Updates a load balancer vip.
# delete_vip            Deletes the specified load balancer vip.
# list_pools            Fetches a list of all load balancer pools for a tenant.
# show_pool             Fetches information of a certain load balancer pool.
# create_pool           Creates a new load balancer pool.
# update_pool           Updates a load balancer pool.
# delete_pool           Deletes the specified load balancer pool.
# retrieve_pool_stats   Retrieves stats for a certain load balancer pool.
# list_members          Fetches a list of all load balancer members for
#                       a tenant.
# show_member           Fetches information of a certain load balancer member.
# create_member         Creates a new load balancer member.
# update_member         Updates a load balancer member.
# delete_member         Deletes the specified load balancer member.
# list_health_monitors  Fetches a list of all load balancer health monitors for
#                       a tenant.
# show_health_monitor   Fetches information of a certain load balancer
#                       health monitor.
# create_health_monitor
#                       Creates a new load balancer health monitor.
# update_health_monitor
#                       Updates a load balancer health monitor.
# delete_health_monitor
#                       Deletes the specified load balancer health monitor.
# associate_health_monitor
#                       Associate  specified load balancer health monitor
#                       and pool.
# disassociate_health_monitor
#                       Disassociate specified load balancer health monitor
#                       and pool.
# create_qos_queue      Creates a new queue.
# list_qos_queues       Fetches a list of all queues for a tenant.
# show_qos_queue        Fetches information of a certain queue.
# delete_qos_queue      Deletes the specified queue.
# list_agents           Fetches agents.
# show_agent            Fetches information of a certain agent.
# update_agent          Updates an agent.
# delete_agent          Deletes the specified agent.
# list_network_gateways
#                       Retrieve network gateways.
# show_network_gateway  Fetch a network gateway.
# create_network_gateway
#                       Create a new network gateway.
# update_network_gateway
#                       Update a network gateway.
# delete_network_gateway
#                       Delete the specified network gateway.
# connect_network_gateway
#                       Connect a network gateway to the specified network.
# disconnect_network_gateway
#                       Disconnect a network from the specified gateway.
# list_gateway_devices  Retrieve gateway devices.
# show_gateway_device   Fetch a gateway device.
# create_gateway_device
#                       Create a new gateway device.
# update_gateway_device
#                       Updates a new gateway device.
# delete_gateway_device
#                       Delete the specified gateway device.
# list_dhcp_agent_hosting_networks
#                       Fetches a list of dhcp agents hosting a network.
# list_networks_on_dhcp_agent
#                       Fetches a list of dhcp agents hosting a network.
# add_network_to_dhcp_agent
#                       Adds a network to dhcp agent.
# remove_network_from_dhcp_agent
#                       Remove a network from dhcp agent.
# list_l3_agent_hosting_routers
#                       Fetches a list of L3 agents hosting a router.
# list_routers_on_l3_agent
#                       Fetches a list of L3 agents hosting a router.
# add_router_to_l3_agent
#                       Adds a router to L3 agent.
# list_firewall_rules   Fetches a list of all firewall rules for a tenant.
# show_firewall_rule    Fetches information of a certain firewall rule.
# create_firewall_rule  Creates a new firewall rule.
# update_firewall_rule  Updates a firewall rule.
# delete_firewall_rule  Deletes the specified firewall rule.
# list_firewall_policies
#                       Fetches a list of all firewall policies for a tenant.
# show_firewall_policy  Fetches information of a certain firewall policy.
# create_firewall_policy
#                       Creates a new firewall policy.
# update_firewall_policy
#                       Updates a firewall policy.
# delete_firewall_policy
#                       Deletes the specified firewall policy.
# firewall_policy_insert_rule
#                       Inserts specified rule into firewall policy.
# firewall_policy_remove_rule
#                       Removes specified rule from firewall policy.
# list_firewalls        Fetches a list of all firewals for a tenant.
# show_firewall         Fetches information of a certain firewall.
# create_firewall       Creates a new firewall.
# update_firewall       Updates a firewall.
# delete_firewall       Deletes the specified firewall.
# remove_router_from_l3_agent
#                       Remove a router from l3 agent.
# get_lbaas_agent_hosting_pool
#                       Fetches a loadbalancer agent hosting a pool.
# list_pools_on_lbaas_agent
#                       Fetches a list of pools hosted by
#                       the loadbalancer agent.
# list_service_providers
#                       Fetches service providers.
# list_credentials      Fetch a list of all credentials for a tenant.
# show_credential       Fetch a credential.
# create_credential     Create a new credential.
# update_credential     Update a credential.
# delete_credential     Delete the specified credential.
# list_network_profile_bindings
#                       Fetch a list of all tenants associated for
#                       a network profile.
# list_network_profiles
#                       Fetch a list of all network profiles for a tenant.
# show_network_profile  Fetch a network profile.
# create_network_profile
#                       Create a network profile.
# update_network_profile
#                       Update a network profile.
# delete_network_profile
#                       Delete the network profile.
# list_policy_profile_bindings
#                       Fetch a list of all tenants associated for
#                       a policy profile.
# list_policy_profiles  Fetch a list of all network profiles for a tenant.
# show_policy_profile   Fetch a network profile.
# update_policy_profile
#                       Update a policy profile.
# create_metering_label
#                       Creates a metering label.
# delete_metering_label
#                       Deletes the specified metering label.
# list_metering_labels  Fetches a list of all metering labels for a tenant.
# show_metering_label   Fetches information of a certain metering label.
# create_metering_label_rule
#                       Creates a metering label rule.
# delete_metering_label_rule
#                       Deletes the specified metering label rule.
# list_metering_label_rules
#                       Fetches a list of all metering label rules for a label.
# show_metering_label_rule
#                       Fetches information of a certain metering label rule.
# list_net_partitions   Fetch a list of all network partitions for a tenant.
# show_net_partition    etch a network partition.
# create_net_partition  Create a network partition.
# delete_net_partition  Delete the network partition.
# create_packet_filter  Create a new packet filter.
# update_packet_filter  Update a packet filter.
# list_packet_filters   Fetch a list of all packet filters for a tenant.
# show_packet_filter    Fetch information of a certain packet filter.
# delete_packet_filter  Delete the specified packet filter.



# Copyright 2010-2021 The pygit2 contributors
#
# This file is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2,
# as published by the Free Software Foundation.
#
# In addition to the permissions in the GNU General Public License,
# the authors give you unlimited permission to link the compiled
# version of this file into combinations with other programs,
# and to distribute those combinations without any restriction
# coming from the use of this file.  (The General Public License
# restrictions do apply in other respects; for example, they cover
# modification of the file, and distribution when not linked into
# a combined executable.)
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING.  If not, write to
# the Free Software Foundation, 51 Franklin Street, Fifth Floor,
# Boston, MA 02110-1301, USA.

"""
Settings mapping.
"""

from ssl import get_default_verify_paths

from . import _pygit2
from ._pygit2 import option
from .errors import GitError


class SearchPathList:

    def __getitem__(self, key):
        return option(_pygit2.GIT_OPT_GET_SEARCH_PATH, key)

    def __setitem__(self, key, value):
        option(_pygit2.GIT_OPT_SET_SEARCH_PATH, key, value)


class Settings:
    """Library-wide settings interface."""

    __slots__ = '_default_tls_verify_paths', '_ssl_cert_dir', '_ssl_cert_file'

    _search_path = SearchPathList()

    def __init__(self):
        """Initialize global pygit2 and libgit2 settings."""
        self._initialize_tls_certificate_locations()

    def _initialize_tls_certificate_locations(self):
        """Set up initial TLS file and directory lookup locations."""
        self._default_tls_verify_paths = get_default_verify_paths()
        try:
            self.set_ssl_cert_locations(
                self._default_tls_verify_paths.cafile,
                self._default_tls_verify_paths.capath,
            )
        except GitError as git_err:
            valid_msg = "TLS backend doesn't support certificate locations"
            if str(git_err) != valid_msg:
                raise
            self._default_tls_verify_paths = None
            self._ssl_cert_file = None
            self._ssl_cert_dir = None

    @property
    def search_path(self):
        """Configuration file search path.

        This behaves like an array whose indices correspond to the
        GIT_CONFIG_LEVEL_* values.  The local search path cannot be
        changed.
        """
        return self._search_path

    @property
    def mwindow_size(self):
        """Get or set the maximum mmap window size"""
        return option(_pygit2.GIT_OPT_GET_MWINDOW_SIZE)

    @mwindow_size.setter
    def mwindow_size(self, value):
        option(_pygit2.GIT_OPT_SET_MWINDOW_SIZE, value)

    @property
    def mwindow_mapped_limit(self):
        """
        Get or set the maximum memory that will be mapped in total by the
        library
        """
        return option(_pygit2.GIT_OPT_GET_MWINDOW_MAPPED_LIMIT)

    @mwindow_mapped_limit.setter
    def mwindow_mapped_limit(self, value):
        return option(_pygit2.GIT_OPT_SET_MWINDOW_MAPPED_LIMIT, value)

    @property
    def cached_memory(self):
        """
        Get the current bytes in cache and the maximum that would be
        allowed in the cache.
        """
        return option(_pygit2.GIT_OPT_GET_CACHED_MEMORY)

    def enable_caching(self, value=True):
        """
        Enable or disable caching completely.

        Because caches are repository-specific, disabling the cache
        cannot immediately clear all cached objects, but each cache will
        be cleared on the next attempt to update anything in it.
        """
        return option(_pygit2.GIT_OPT_ENABLE_CACHING, value)

    def disable_pack_keep_file_checks(self, value=True):
        """
        This will cause .keep file existence checks to be skipped when
        accessing packfiles, which can help performance with remote
        filesystems.
        """
        return option(_pygit2.GIT_OPT_DISABLE_PACK_KEEP_FILE_CHECKS, value)

    def cache_max_size(self, value):
        """
        Set the maximum total data size that will be cached in memory
        across all repositories before libgit2 starts evicting objects
        from the cache.  This is a soft limit, in that the library might
        briefly exceed it, but will start aggressively evicting objects
        from cache when that happens.  The default cache size is 256MB.
        """
        return option(_pygit2.GIT_OPT_SET_CACHE_MAX_SIZE, value)

    def cache_object_limit(self, object_type, value):
        """
        Set the maximum data size for the given type of object to be
        considered eligible for caching in memory.  Setting to value to
        zero means that that type of object will not be cached.
        Defaults to 0 for GIT_OBJECT_BLOB (i.e. won't cache blobs) and 4k
        for GIT_OBJECT_COMMIT, GIT_OBJECT_TREE, and GIT_OBJECT_TAG.
        """
        return option(_pygit2.GIT_OPT_SET_CACHE_OBJECT_LIMIT, object_type, value)

    @property
    def ssl_cert_file(self):
        """TLS certificate file path."""
        return self._ssl_cert_file

    @ssl_cert_file.setter
    def ssl_cert_file(self, value):
        """Set the TLS cert file path."""
        self.set_ssl_cert_locations(value, self._ssl_cert_dir)

    @ssl_cert_file.deleter
    def ssl_cert_file(self):
        """Reset the TLS cert file path."""
        self.ssl_cert_file = self._default_tls_verify_paths.cafile

    @property
    def ssl_cert_dir(self):
        """TLS certificates lookup directory path."""
        return self._ssl_cert_dir

    @ssl_cert_dir.setter
    def ssl_cert_dir(self, value):
        """Set the TLS certificate lookup folder."""
        self.set_ssl_cert_locations(self._ssl_cert_file, value)

    @ssl_cert_dir.deleter
    def ssl_cert_dir(self):
        """Reset the TLS certificate lookup folder."""
        self.ssl_cert_dir = self._default_tls_verify_paths.capath

    def set_ssl_cert_locations(self, cert_file, cert_dir):
        """
        Set the SSL certificate-authority locations.

        - `cert_file` is the location of a file containing several
          certificates concatenated together.
        - `cert_dir` is the location of a directory holding several
          certificates, one per file.

        Either parameter may be `NULL`, but not both.
        """
        option(_pygit2.GIT_OPT_SET_SSL_CERT_LOCATIONS, cert_file, cert_dir)
        self._ssl_cert_file = cert_file
        self._ssl_cert_dir = cert_dir



"""
Module for sending messages to Pushbullet (https://www.pushbullet.com)

.. versionadded:: 2015.8.0

Requires an ``api_key`` in ``/etc/salt/minion``:

.. code-block:: yaml

    pushbullet:
      api_key: 'ABC123abc123ABC123abc123ABC123ab'

For example:

.. code-block:: yaml

    pushbullet:
      device: "Chrome"
      title: "Example push message"
      body: "Message body."

"""

import logging

try:
    import pushbullet

    HAS_PUSHBULLET = True
except ImportError:
    HAS_PUSHBULLET = False

log = logging.getLogger(__name__)


def __virtual__():
    if not HAS_PUSHBULLET:
        return (False, "Missing pushbullet library.")
    if not __salt__["config.get"]("pushbullet.api_key") and not __salt__["config.get"](
        "pushbullet:api_key"
    ):
        return (False, "Pushbullet API Key Unavailable, not loading.")
    return True


class _SaltPushbullet:
    def __init__(self, device_name):
        api_key = __salt__["config.get"]("pushbullet.api_key") or __salt__[
            "config.get"
        ]("pushbullet:api_key")
        self.pb = pushbullet.Pushbullet(api_key)
        self.target = self._find_device_by_name(device_name)

    def push_note(self, title, body):
        push = self.pb.push_note(title, body, device=self.target)
        return push

    def _find_device_by_name(self, name):
        for dev in self.pb.devices:
            if dev.nickname == name:
                return dev


def push_note(device=None, title=None, body=None):
    """
    Pushing a text note.

    :param device:   Pushbullet target device
    :param title:    Note title
    :param body:     Note body

    :return:            Boolean if message was sent successfully.

    CLI Example:

    .. code-block:: bash

        salt "*" pushbullet.push_note device="Chrome" title="Example title" body="Example body."
    """
    spb = _SaltPushbullet(device)
    res = spb.push_note(title, body)

    return res
